/*
 * Copyright (C) 2003 Maxim Stepin ( maxst@hiend3d.com )
 *
 * Copyright (C) 2010 Cameron Zemek ( grom@zeminvaders.net)
 *
 * Copyright (C) 2011 Francois Gannaz <mytskine@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef __HQX_COMMON_H_
#define __HQX_COMMON_H_

#include <stdlib.h>
#include <stdint.h>

#define ENABLE_ALPHA

#define Ymask 0x00FF0000
#define Umask 0x0000FF00
#define Vmask 0x000000FF
#define trY   0x00300000
#define trU   0x00000700
#define trV   0x00000006

/* 32 bits color */
#define COLOR_UPPER_BOUND 0x1000000
#define MASK_2  0x00FF00
#define MASK_13 0xFF00FF
#define MASK_ALPHA 0xFF000000
#define MASK_NOALPHA 0x00FFFFFF
typedef uint32_t color_t;

/* barycenter of 2 colors : weigth1, weigth2, shift (i.e. log2(w1+w2)) */
#ifdef ENABLE_ALPHA
#define BAR2(w1,w2,s) \
    (c1 == c2) ? c1 : \
    (((((c1 & MASK_ALPHA) >> 24) * w1 + ((c2 & MASK_ALPHA) >> 24) * w2) << (24-s)) & MASK_ALPHA) + \
    ((((c1 & MASK_2) * w1 + (c2 & MASK_2) * w2) >> s) & MASK_2)	\
    + ((((c1 & MASK_13) * w1 + (c2 & MASK_13) * w2) >> s) & MASK_13)
#else
#define BAR2(w1,w2,s) \
    (c1 == c2) ? c1 : \
    ((((c1 & MASK_2) * w1 + (c2 & MASK_2) * w2) >> s) & MASK_2)	\
    + ((((c1 & MASK_13) * w1 + (c2 & MASK_13) * w2) >> s) & MASK_13)
#endif

/* barycenter of 3 colors */
#ifdef ENABLE_ALPHA
#define BAR3(w1,w2,w3,s) \
    (((((c1 & MASK_ALPHA) >> 24) * w1 + ((c2 & MASK_ALPHA) >> 24) * w2 + ((c3 & MASK_ALPHA) >> 24) * w3) << (24-s)) & MASK_ALPHA) + \
    ((((c1 & MASK_2) * w1 + (c2 & MASK_2) * w2 + (c3 & MASK_2) * w3) >> s) & MASK_2) \
    + ((((c1 & MASK_13) * w1 + (c2 & MASK_13) * w2 + (c3 & MASK_13) * w3) >> s) & MASK_13)
#else
#define BAR3(w1,w2,w3,s) \
    ((((c1 & MASK_2) * w1 + (c2 & MASK_2) * w2 + (c3 & MASK_2) * w3) >> s) & MASK_2) \
    + ((((c1 & MASK_13) * w1 + (c2 & MASK_13) * w2 + (c3 & MASK_13) * w3) >> s) & MASK_13)
#endif

/* RGB to YUV lookup table */
uint32_t   RGBtoYUV[COLOR_UPPER_BOUND];
uint32_t   YUV1, YUV2;

static inline uint32_t rgb_to_yuv(color_t c)
{
    return RGBtoYUV[MASK_NOALPHA & c];
}

/* Test if there is difference in color */
static inline int Diff(color_t w1, color_t w2)
{
    // Mask against RGB_MASK to discard the alpha channel
    YUV1 = rgb_to_yuv(w1);
    YUV2 = rgb_to_yuv(w2);
    return ( ( abs((YUV1 & Ymask) - (YUV2 & Ymask)) > trY ) ||
            ( abs((YUV1 & Umask) - (YUV2 & Umask)) > trU ) ||
            ( abs((YUV1 & Vmask) - (YUV2 & Vmask)) > trV ) );
}

/* Interpolate functions */

static inline color_t Interp1(color_t c1, color_t c2)
{
    //return (c1*3+c2) >> 2;
	return BAR2(3,1,2);
}

static inline color_t Interp2(color_t c1, color_t c2, color_t c3)
{
    //return (c1*2+c2+c3) >> 2;
    return BAR3(2,1,1,2);
}

static inline color_t Interp3(color_t c1, color_t c2)
{
    //return (c1*7+c2)/8;
    return BAR2(7,1,3);
}

static inline color_t Interp4(color_t c1, color_t c2, color_t c3)
{
    //return (c1*2+(c2+c3)*7)/16;
    return BAR3(2,7,7,4);
}

static inline color_t Interp5(color_t c1, color_t c2)
{
    //return (c1+c2) >> 1;
    return BAR2(1,1,1);
}

static inline color_t Interp6(color_t c1, color_t c2, color_t c3)
{
    //return (c1*5+c2*2+c3)/8;
    return BAR3(5,2,1,3);
}

static inline color_t Interp7(color_t c1, color_t c2, color_t c3)
{
    //return (c1*6+c2+c3)/8;
    return BAR3(6,1,1,3);
}

static inline color_t Interp8(color_t c1, color_t c2)
{
    //return (c1*5+c2*3)/8;
    return BAR2(5,3,3);
}

static inline color_t Interp9(color_t c1, color_t c2, color_t c3)
{
    //return (c1*2+(c2+c3)*3)/8;
    return BAR3(2,3,3,3);
}

static inline color_t Interp10(color_t c1, color_t c2, color_t c3)
{
    //return (c1*14+c2+c3)/16;
    return BAR3(14,1,1,4);
}

#endif
